// last modified Nov/11/99 by Fabien Costantini
//
// new features : 
//   new config option : settings for make tool
//   go option in search window + dbl clicks handled
//   resizable search window
//   now compile with visualc++ 6 under win32

#if defined(WIN32)
#define popen _popen
#endif

 
#include <FL/Fl_ToggleTree.H>
#include <FL/Fl_Window.H>
#include <FL/Fl_Button.H>
#include <FL/Fl_Radio_Button.H>
#include <FL/Fl_Scroll.H>
#include <FL/Fl.H>

#include <FL/folder_small.xpm>
#include <FL/file_small.xpm>
#include "function_small.xpm"
#include "point_small.xpm"
#include "box_small.xpm"
#include "boxes_small.xpm"
#include "global_small.xpm"

#include "TagsGUI.H"
#include "Fl_TTreeResizer.H"
#include "settingparser.hh"

#include <stdio.h>
#include <stdlib.h>

const int MAX_SEARCH_ENTRY = 25;

Fl_Pixmap* fileSmall;
Fl_Pixmap* folderSmall;
Fl_Pixmap* pointSmall;
Fl_Pixmap* functionSmall;
Fl_Pixmap* boxSmall;
Fl_Pixmap* boxesSmall;
Fl_Pixmap* globalSmall;

class TagNodeInfo
{
public:
	TagNodeInfo(int t)
	{
		type=t;
	}
	int type;
};

#define kTagNodeGlobal	1
#define kTagNodeGroup	2
#define kTagNodeMembers	10
#define kTagNodeMethods 11
#define kTagNodeSource  12
#define kTagNodeRef		20
#define kTagNodeCategory	50
#define kTagNodeDeleting	128
#define kTagNodeSources	        256

class GroupNodeInfo:public TagNodeInfo
{
public:
	GroupNodeInfo():TagNodeInfo(kTagNodeGroup){
		methods = 0;
		members = 0;
	}
	Fl_ToggleNode* methods;
	Fl_ToggleNode* members;
};

// addon F.C
class SourceNodeInfo:public TagNodeInfo
{
public:
	SourceNodeInfo():TagNodeInfo(kTagNodeSource){}
};
// end F.C

class MembersNodeInfo:public TagNodeInfo
{
public:
	MembersNodeInfo():TagNodeInfo(kTagNodeMembers){}
};

class MethodsNodeInfo:public TagNodeInfo
{
public:
	MethodsNodeInfo():TagNodeInfo(kTagNodeMethods){}
};

class RefNodeInfo:public TagNodeInfo
{
public:
	RefNodeInfo(char* Filename,int Line):TagNodeInfo(kTagNodeRef){
		filename=Filename;
		line=Line;
	}
	char* filename;
	int line;
};

void TagsGUI::SortByLine(void)
{
	tree->sort_tree(LineSortFunction);
	tree->redraw();
}

void TagsGUI::SortByName(void)
{
	tree->sort_tree(NameSortFunction);
	tree->redraw();
}

int TagsGUI::LineSortFunction(Fl_Node* a, Fl_Node* b) {
	TagNodeInfo* ia=(TagNodeInfo*) ((Fl_ToggleNode*)a)->data();
	TagNodeInfo* ib=(TagNodeInfo*) ((Fl_ToggleNode*)b)->data();
	
	if (ia->type==kTagNodeGlobal) return -1;
	if (ib->type==kTagNodeGlobal) return 1;
	if (ia->type==kTagNodeMembers) return -1;
	if (ib->type==kTagNodeMembers) return 1;
		
	if (ia->type==kTagNodeRef) {
		RefNodeInfo* ra=(RefNodeInfo*) ia;
		RefNodeInfo* rb=(RefNodeInfo*) ib;
		int c=strcmp(rb->filename,ra->filename);
		if (c==0) {
			if (ra->line<rb->line) return -1;
			if (ra->line>rb->line) return 1;
			return 0;
		}
		return c;
	}
	if (ia->type>kTagNodeCategory) {
		if (ia->type>ib->type) return 1;
		if (ib->type>ia->type) return -1;
		return 0;
	}		
		
	return strcmp(((Fl_ToggleNode*)a)->label(), ((Fl_ToggleNode*)b)->label());
}

int TagsGUI::NameSortFunction(Fl_Node* a, Fl_Node* b) {
	TagNodeInfo* ia=(TagNodeInfo*) ((Fl_ToggleNode*)a)->data();
	TagNodeInfo* ib=(TagNodeInfo*) ((Fl_ToggleNode*)b)->data();
	
	if (ia->type==kTagNodeGlobal) return -1;
	if (ib->type==kTagNodeGlobal) return 1;
	if (ia->type==kTagNodeMembers) return -1;
	if (ib->type==kTagNodeMembers) return 1;
	if (ia->type>kTagNodeCategory) {
		if (ia->type>ib->type) return 1;
		if (ib->type>ia->type) return -1;
		return 0;
	}		
	return strcmp(((Fl_ToggleNode*)a)->label(), ((Fl_ToggleNode*)b)->label());
}

Fl_ToggleNode* catnodes[8];
int ncats=0;

class TagTreeLoader
{
public:
	Fl_ToggleTree* tree;
	char** filenames;
	int nfilenames;		
	TagTreeLoader(Fl_ToggleTree* t)
	{
		tree=t;
		filenames=0;
		nfilenames=0;
	}

	Fl_ToggleNode* new_group(char* str,Fl_ToggleNode* cat,Fl_Pixmap* pixmap=folderSmall)
	{
		tree->traverse_start(cat);

		Fl_ToggleNode* n=new Fl_ToggleNode(str,1,pixmap);
		GroupNodeInfo* i = new GroupNodeInfo;
		tree->add_sub(n);
		tree->close(n);
		n->data(i);
		if (strcmp(cat->label(),"class")==0) {
			i->methods=tree->add_sub("Methods",1,boxesSmall);
			MethodsNodeInfo* i2 = new MethodsNodeInfo;
			i->methods->data(i2);
			tree->close(i->methods);
			i->members=tree->add_next("Members",1,boxesSmall);
			MembersNodeInfo* i3 = new MembersNodeInfo;
			i->members->data(i3);
			tree->close(i->members);
		}else{
			i->members=n;
			i->methods=n;
		}
		
		return n;
	};

	void new_global(void)
	{
		if (get_cat("global")) return;
		Fl_ToggleNode* global=new_cat("global",globalSmall);
		GroupNodeInfo* i = new GroupNodeInfo;
		i->type=1;
		global->data(i);

		i->methods=tree->add_sub("Functions",1,boxesSmall);
		MethodsNodeInfo* i2 = new MethodsNodeInfo;
		i->methods->data(i2);
		tree->close(i->methods);
		i->members=tree->add_next("Variables",1,boxesSmall);
		MembersNodeInfo* i3 = new MembersNodeInfo;
		i->members->data(i3);
		tree->close(i->members);
	};
    // F.C add-on
        void new_sources(void)
	{
	    static GroupNodeInfo* i = new GroupNodeInfo;
		if (!get_cat("sources"))
		{
			Fl_ToggleNode* sources=new_cat("sources",folderSmall);
			i->type= kTagNodeSources;
			sources->data(i);
			i->methods=tree->add_sub("Include",1,folderSmall);
			SourceNodeInfo* i2 = new SourceNodeInfo;
			i->methods->data(i2);
			tree->close(i->methods);
			i->members=tree->add_next("Source",1,folderSmall);
			SourceNodeInfo* i3 = new SourceNodeInfo;
			i->members->data(i3);
			tree->close(i->members);
		}
		Fl_ToggleNode* node = NULL;
		for (int n=0; n<nfilenames;n++)
		{
		    char * s = name_only(filenames[n]);
		    size_t l = strlen(s);
		    if (s && l>2)
		    {
			// is it an std include file ?
			if (strcmp(s+l-2,".H")==0 || strcmp(s+l-2,".h")==0 
			    || strcmp(s+l-3,".hh")==0 || strcmp(s+l-4,".hpp")==0 )
			    node = i->methods;
			else
			    node = i->members;
			tree->traverse_start(node);
			Fl_ToggleNode* nn=new Fl_ToggleNode(s,0,fileSmall);
			tree->add_sub(nn);
			RefNodeInfo* ni=new RefNodeInfo(filenames[n],0);
			nn->data(ni);
		    }
		}
	};

	Fl_ToggleNode* get_cat(char* catstr) {
		int cati;
		
		for (cati=0;cati<ncats;cati++) {
			if (strcmp(catstr,catnodes[cati]->label())==0) {
				return catnodes[cati];
			}
		}
		return 0;	
	}

	Fl_ToggleNode* new_cat(char* catstr,Fl_Pixmap* pixmap=folderSmall)
	{
		Fl_ToggleNode* cat;
		if (ncats) {
			tree->traverse_start(catnodes[ncats-1]);
		}else{
			tree->traverse_start();
		}
		cat=new Fl_ToggleNode(catstr,1,pixmap);
		tree->add_next(cat);
		GroupNodeInfo* i = new GroupNodeInfo;
		i->type=kTagNodeCategory+ncats;
		cat->data(i);
		tree->close(cat);		
		catnodes[ncats++]=cat;
	
		return cat;
	}
	
	Fl_ToggleNode* get_or_new_cat(char* catstr)
	{
		Fl_ToggleNode* cat=get_cat(catstr);
		if (cat) return cat;
		return new_cat(catstr);	
	}
	
	void load(char** files,int nfiles)
	{
		new_global();
	
		get_or_new_cat("class");
		get_or_new_cat("struct");
		get_or_new_cat("enum");
		
		char* cmd=0;
		int i;
		if (nfiles) {
			int l=0;
			for (i=0;i<nfiles;i++) {
				l+=1+strlen(files[i]);
			}
			char* cmdctags="ctags -n -i cdefgmpstuv -f -";
			l+=strlen(cmdctags);
			cmd=new char[l+1];
			strcpy(cmd,cmdctags);
			for (i=0;i<nfiles;i++) {
				strcat(cmd," ");
				strcat(cmd,files[i]);
			}
		}
		
		char line[256];
		FILE* ctagsout=0;
		
		if (nfiles) {
#ifdef VERBOSE
			printf("running %s\n",cmd);
#endif
			ctagsout=popen(cmd,"r");			
			delete cmd;
		}else{
#ifdef VERBOSE
			printf("opening file \"tags\"");
#endif
			ctagsout=fopen("tags","r");			
		}
		if (ctagsout) while (fgets(line,255,ctagsout)) {
			char* name=strtok(line,"\t");
			char* file=strtok(0,"\t");
			char* line=strtok(0,"\t");
			char* type=strtok(0,"\t");
			char* tmp1=strtok(0,"\n");
			char* owner=0;
			Fl_ToggleNode* cat=catnodes[0];
						
			{
			    char* catstr=strtok(tmp1,":");
			    owner=strtok(0,"\0");
			    if (catstr) {
				cat=get_or_new_cat(catstr);
				if (owner==0) {
				    owner="unknown";
				}
			    }									
			}

			if (type && (type[0]=='f' || type[0]=='m' || type[0]=='e')) {
			    Fl_ToggleNode * curr=0;
				
			    if (owner) {
				tree->traverse_start(cat);
				curr = tree->traverse_forward();
	  			while (curr) {
				    TagNodeInfo* ni=(TagNodeInfo*) curr->data();
				    if (ni->type==2 || ni->type==(kTagNodeGroup|kTagNodeDeleting)) {
					if (strcmp(curr->label(),owner) == 0) {
					    ni->type=kTagNodeGroup;
					    break;
					}
				    }
				    if (ni->type==kTagNodeCategory) {
					curr=0;
				    }else{
		    			curr = tree->traverse_forward();
				    }
				}
			    }
			    if (curr == 0) {
				if (owner) {
				    curr = new_group(owner,cat,boxSmall);
				}else {
				    curr = catnodes[0];
				}
			    }
			    if (type[0]=='f') {
				new_method((GroupNodeInfo*) curr->data(),
					   name,file,atoi(line));
			    }else{
				new_member((GroupNodeInfo*) curr->data(),
					   name,file,atoi(line));
			    }
			}	

		}
		// add sources at the end...
		new_sources();
	}
	
    char * name_only(char *filename)
    {
	if (filename )
	{
	    char *s;
	    size_t l = strlen(filename);
	    if (l)
		for (s=filename+l-1;s>=filename;s--) 
		    if (*s == '/' || *s =='\\' || *s==':') return (s+1);
	}
	return filename;
    }

    char* filenamestr(char* filename)
	{
	    for (int i=0;i<nfilenames;i++) {
		if (strcmp(filename,filenames[i])==0) return filenames[i];
		}
		
		filenames=(char**) realloc(filenames,(nfilenames+1)*sizeof(char*));
		filename=strdup(filename);
		filenames[nfilenames++]=filename;
		return filename;
	}
	void new_method(GroupNodeInfo* ci,char* str,char* filename,int line)
	{
		tree->traverse_start(ci->methods);
		Fl_ToggleNode* n=new Fl_ToggleNode(str,0,functionSmall);
		tree->add_sub(n);
		RefNodeInfo* ni=new RefNodeInfo(filenamestr(filename),line);
		n->data(ni);
	}
	void new_member(GroupNodeInfo* ci,char* str,char* filename,int line)
	{
		tree->traverse_start(ci->members);
		Fl_ToggleNode* n=new Fl_ToggleNode(str,0,pointSmall);
		tree->add_sub(n);
		RefNodeInfo* ni=new RefNodeInfo(filenamestr(filename),line);
		n->data(ni);
	}
};

void TagsGUI::cb_tree(Fl_Widget* t) {
	((TagsGUI*) (t->parent()->parent()->user_data()))->TreeSelect();
}

void TagsGUI::TreeSelect(void) {
  if (tree->selected()) {
	TagNodeInfo* ti=(TagNodeInfo*) tree->selected()->data();
	if (ti->type==kTagNodeRef) {
		RefNodeInfo* ri=(RefNodeInfo*) ti;
        char cmd[1024];
		int i=0,j=0;
		while (invoke_editor_str[i]) {
			if (invoke_editor_str[i]=='%') {
				i++;
				switch (invoke_editor_str[i]) {
				case '%':
					cmd[j]='%';
					break;
				case 'f':
					cmd[j]='\0';
					strcat(&cmd[j],ri->filename);
					j+=strlen(ri->filename);
					break;
				case 'n':
					cmd[j]='\0';
					{
						char tmp[8];
						sprintf(tmp,"%d",ri->line);
						strcat(&cmd[j],tmp);
						j+=strlen(tmp);
					}
					break;
				default:
					cmd[j++]='%';
					cmd[j]=invoke_editor_str[i];
				}
			}else{
				cmd[j]=invoke_editor_str[i];
				j++;
			}		
			i++;
		}
		cmd[j]='\0';
#ifdef VERBOSE
		printf("calling editor: %s %s\n",invoke_editor_str,cmd);
#endif
		system(cmd);
	}
  }
}

#define kByName 0

char prefsfilename[1024];

void TagsGUI::Init(void)
{
	sprintf(prefsfilename,"%s/.fltags",getenv("HOME"));

	mSettingsParser.Add(
		"make_ctags",&make_ctags_str,
			strdup("make ctags"));
	mSettingsParser.Add(
		"invoke_ctags",&invoke_ctags_str,
			strdup("ctags -n -i cdefgmpstuv -f -"));
	mSettingsParser.Add(
		"invoke_editor",&invoke_editor_str,
			strdup("nc -noask +%n %f"));
	mSettingsParser.Add(
		"sort_type",&sorttype,0);

	mSettingsParser.Parse(prefsfilename);

	InitTree();
}

void TagsGUI::Load(char** files,int nfiles)
{
	files_ = files;
	nfiles_ = nfiles;
	
	_Load(files_,nfiles_);
		
}

void TagsGUI::_Load(char** files,int nfiles)
{
    TagTreeLoader tl(tree);
  
    tl.load(files,nfiles);

    if (sorttype==kByName)
	SortByName();
    else
	SortByLine();	
}

void TagsGUI::MakeTagsAndReload(void)
{
	system(make_ctags_str);
	
	Reload(1);
}

void TagsGUI::Reload(int force_load_tags)
{
	Fl_ToggleNode *curr;
	curr=(Fl_ToggleNode*) tree->traverse_start();
	
	while (curr) {
		TagNodeInfo* ti=(TagNodeInfo*) curr->data();
		if (ti->type==kTagNodeGroup) ti->type=kTagNodeGroup|kTagNodeDeleting;
		else if (ti->type==kTagNodeRef) tree->remove(curr);
		curr=(Fl_ToggleNode*) tree->traverse_forward();
	}

	if (force_load_tags) {
		_Load(0,0);
	}else{
		_Load(files_,nfiles_);
	}

	curr=(Fl_ToggleNode*) tree->traverse_start();
	
	while (curr) {
		TagNodeInfo* ti=(TagNodeInfo*) curr->data();

		if (ti->type==(kTagNodeGroup|kTagNodeDeleting)){
			tree->remove(curr);
		}
		curr=(Fl_ToggleNode*) tree->traverse_forward();
	}

	tree->unselect();

	scroll->redraw();
}

void TagsGUI::InitTree(void)
{
	tree = new Fl_ToggleTree(
		scroll->x()+2,
		scroll->y()+2,
		scroll->w()-scroll->scrollbar.w()-3, 0); 
	scroll->add(tree);

	Fl_ToggleTreeResizer* ttr = new Fl_ToggleTreeResizer(scroll,tree);

	((Fl_Group*)(scroll->parent()))->add(ttr);

	tree->callback(cb_tree);
}

void TagsGUI::EditPreferences(void)
{
	int prevsorttype=sorttype;
	
	window->deactivate();
	PreferencesGUI p(this);
	p.Run();
	window->activate();

	if (sorttype!=prevsorttype) {
		if (sorttype==kByName)
			SortByName();
		else
			SortByLine();	
	}
}

void TagsGUI::SavePreferences(void)
{
	mSettingsParser.Save(prefsfilename);
}

void TagsGUI::About(void)
{
	window->deactivate();
	AboutBox about;
	about.Run();
	window->activate();
}

void TagsGUI::Run(void)
{
	window->resizable(scroll);
	window->show();
	while (!done /*&& window->visible()*/) {
		Fl::wait();
	}
	delete window;	
}

void TagsGUI::Quit(void)
{
	done=1;
}

void TagsGUI::Search(void)
{
	SearchGUI s(this);
	s.Run();
}

int CalcScore(const char* str1,const char* str2)
{
	int i=0;
	int score=0;
	int a=20;
	while (str1[i] && str2[i]) {
		int dif=str1[i]-str2[i];
		if (dif<0) dif=-dif;
		score+=((127-dif)<<a);
		a-=4;
		i++;
	} 
	return score;
}


void SearchGUI::Search(void)
{
	Fl_ToggleTree* tree=mGUI->Tree();
	Fl_ToggleNode* curr;
	
	curr=(Fl_ToggleNode*) tree->traverse_start();

	mResultsTree->traverse_start();
	while (mResultsTree->first()) mResultsTree->clear();
	mResultsTree->traverse_start();
	
	Fl_ToggleNode* best[MAX_SEARCH_ENTRY];
	int score[MAX_SEARCH_ENTRY];
	int i;
	
	for (i=0;i<MAX_SEARCH_ENTRY;i++) {
		best[i]=0;
		score[i]=0;
	}
	
	while (curr) {
		TagNodeInfo* ti=(TagNodeInfo*) curr->data();

		if (ti->type==(kTagNodeRef)){
			int curscore=CalcScore(mSearchTerm->value(),curr->label());
			
			i=MAX_SEARCH_ENTRY-1;
			while (i>=0 && curscore>score[i]) {
				i--;
			}
			i++;
			if (i<MAX_SEARCH_ENTRY) {
				int j;
				for (j=MAX_SEARCH_ENTRY-1;j>i;j--) {
					score[j]=score[j-1];
					best[j]=best[j-1];
				}
				score[i]=curscore;
				best[i]=curr;
			}
		}
		curr=(Fl_ToggleNode*) tree->traverse_forward();
	}

	for (i=0;i<MAX_SEARCH_ENTRY;i++) {
		if (best[i]) {
			char str[256];
			strcpy(str,best[i]->label());
			tree->traverse_start(best[i]);
			tree->traverse_up();
			curr=tree->current();
			if (curr) {
				TagNodeInfo* ti=(TagNodeInfo*) curr->data();
				if (ti->type == kTagNodeMethods ||  ti->type == kTagNodeMembers) {
					tree->traverse_up();
					curr=tree->current();
					TagNodeInfo* ti=(TagNodeInfo*) curr->data();
					if (ti->type != kTagNodeGlobal) {
						strcpy(str,curr->label());
						strcat(str,"::");
						strcat(str,best[i]->label());
					}
				}else{
					if (ti->type != kTagNodeGlobal) {
						strcpy(str,curr->label());
						strcat(str,"::");
						strcat(str,best[i]->label());
					}
				}
			}
			Fl_ToggleNode* node;
			mResultsTree->add_next(node=new Fl_ToggleNode(str,0,
				best[i]->pixmap()));
			node->data(best[i]);
		}
	}

	mResultsTree->update_height();
}

void SearchGUI::cb_tree(Fl_Widget* t) {
	((SearchGUI*) (t->parent()->parent()->user_data()))->TreeSelect();
}

void SearchGUI::TreeSelect(void) {
	Fl_ToggleTree* tree=mGUI->Tree();
	
  if (mResultsTree->selected()) {
		tree->traverse_start((Fl_ToggleNode*) mResultsTree->selected()->data());
		Fl_ToggleNode* curr=tree->current();
		while (curr) {
			printf("%s\n",curr->label());
			fflush(stdout);
			tree->open(curr);
			tree->traverse_up();
			if (curr==tree->current()) curr=0;
			else curr=tree->current();
		}
		tree->update_height();
		Fl_ToggleNode* s=(Fl_ToggleNode*) mResultsTree->selected()->data();
		int y=tree->ypos(s);
		if (
			y+tree->y() > mGUI->scroll->y()+mGUI->scroll->h()-4 ||
			y+tree->y() < mGUI->scroll->y() 
		) {
			mGUI->scroll->position(0,y-mGUI->scroll->h()/2);
			tree->select(s);
			mGUI->scroll->redraw();
		}else{
			tree->select(s);
			mGUI->scroll->redraw();
		}
		// F.C add-on : dbl-click go there directly
		if (Fl::event_clicks()) Go();
	}
}

// add-on F.C
void SearchGUI::Go(void)
{
    if (mGUI) 
    {
      mGUI->tree->selection();
      mGUI->TreeSelect();
    }
}
void SearchGUI::Run(void)
{
  	mResultsTree->resize(
  		mResultsScroll->x()+2,
  		mResultsScroll->y()+2,
  		mResultsScroll->w()-4, 0); 

	mResultsTree->callback(cb_tree);

	//mResultsScroll->add(mResultsTree);

	done=0;
	window->show();
	while (!done && window->visible()) {
		Fl::wait();
	}
	delete window;
}

void SearchGUI::Done(void)
{
	done=1;
}

void PreferencesGUI::Help(void)
{
	HelpGUI g;
	g.Run();
}

void TagsGUI::Help(void)
{
	HelpGUI g;
	g.Run();
}

void PreferencesGUI::Run(void)
{
	mSortByName->value(mGUI->sorttype==0);
	mSortByLine->value(mGUI->sorttype==1);
	mCtagsMake->value(mGUI->make_ctags_str);
	mCtagsInvoke->value(mGUI->invoke_ctags_str);
	mEditorInvoke->value(mGUI->invoke_editor_str);

	window->show();
	while (!done && window->visible()) {
		Fl::wait();
	}
	delete window;
}

void PreferencesGUI::OK(void)
{
	mGUI->sorttype=mSortByName->value()?0:1;
	delete mGUI->make_ctags_str;
	mGUI->make_ctags_str=strdup(mCtagsMake->value());
	delete mGUI->invoke_ctags_str;
	mGUI->invoke_ctags_str=strdup(mCtagsInvoke->value());
	delete mGUI->invoke_editor_str;
	mGUI->invoke_editor_str=strdup(mEditorInvoke->value());
	done=1;
}

void PreferencesGUI::Cancel(void)
{
	done=1;
}

void AboutBox::Run(void)
{
	done=0;
	window->show();
	while (!done && window->visible()) {
		Fl::wait();
	}
	delete window;
}

void AboutBox::Done(void)
{
	done=1;
}

void HelpGUI::Run(void)
{
	done=0;
	window->show();
	while (!done && window->visible()) {
		Fl::wait();
	}
	delete window;
}

void HelpGUI::Done(void)
{
	done=1;
}


main(int argc,char** argv) 
{
  TagsGUI gui;

  fileSmall = new Fl_Pixmap(file_small);
  folderSmall = new Fl_Pixmap(folder_small);
  pointSmall = new Fl_Pixmap(point_small);
  functionSmall = new Fl_Pixmap(function_small);
 	boxSmall = new Fl_Pixmap(box_small);
  boxesSmall = new Fl_Pixmap(boxes_small);
  globalSmall = new Fl_Pixmap(global_small);

  gui.Init();
  gui.Load(argv+1,argc-1);

  gui.Run();

  return 0;
}

