#include "settingparser.hh"
#include <string.h>
#include <stdlib.h>
#include <fstream.h>

#ifndef null
#define null (0L)
#endif

SettingsParser::SettingsParser()
{
	first=last=null;
}

void SettingsParser::Add(Setting* setting)
{
	if (first == null) {
		first = last = setting;
	} else {
		last->next = setting;
		last = setting;
	}
}

void SettingsParser::Add(char* key,bool*  valptr,bool def)
{
	Setting* s=new Setting;
	s->key  = key;
	s->data = (void*) valptr;
	s->typ  = 0;
	s->next = 0;
	*valptr = def;
	
	Add(s);
}

void SettingsParser::Add(char* key,char** valptr,char* def)
{
	Setting* s=new Setting;
	s->key  = key;
	s->data = (void*) valptr;
	s->typ  = 1;
	s->next = 0;
	*valptr = def;

	Add(s);
}

void SettingsParser::Add(char* key,int* valptr,int def)
{
	Setting* s=new Setting;
	s->key  = key;
	s->data = (void*) valptr;
	s->typ  = 2;
	s->next = 0;
	*valptr = def;

	Add(s);
}

void SettingsParser::Add(char* key,float* valptr,float def)
{
	Setting* s=new Setting;
	s->key  = key;
	s->data = (void*) valptr;
	s->typ  = 3;
	s->next = 0;
	*valptr = def;

	Add(s);
}


void SettingsParser::SetFlag(bool* b,char* str)
{
	if (strcmp(str,"0")==0 || strcmp(str,"false")==0){
		*b=0;
		return;
	}
	if (strcmp(str,"1")==0 || strcmp(str,"true")==0){
		*b=1;
		return;
	}
	mError|=1;
}

void SettingsParser::SetString(char** ptr,char* str)
{
	*ptr=strdup(str);
}

void SettingsParser::SetInt(int* ptr,char* str)
{
	*ptr=atoi(str);
}

void SettingsParser::SetFloat(float* ptr,char* str)
{
	*ptr=atof(str);
}


void SettingsParser::Set(char* key,char* val) {
	Setting* s=first;
	mFound = 0;
	while (s) {
		if (strcmp(key,s->key)==0) {
			mFound = 1;
			switch (s->typ) {
				case 0:
					SetFlag((bool*) s->data,val);
					return;
				case 1:
					SetString((char**) s->data,val);					 
					return;
				case 2:
					SetInt((int*) s->data,val);
					return;
				case 3:
					SetFloat((float*) s->data,val);
					return;
			}	
		}
		s=s->next;
	}
}

void SettingsParser::Save(char* filename)
{
	ofstream settings(filename);
	Setting* s=first;
	while (s) {
//		cout << s->key << "\n";
		settings << s->key << " = ";
		switch (s->typ) {
			case 0:
				settings << "\"" << (*(bool*) s->data) << "\"";
				break;
			case 1:
				settings << *(char**) s->data;
				break;
			case 2:
				settings << *(int*) s->data;
				break;
			case 3:
				settings << *(float*) s->data;
				break;
		}	
		settings << "\n";
		s=s->next;
	}
}

void SettingsParser::Parse(char* filename)
{
	int line=0;	

	ifstream settings(filename);
	
	mHasErrors = false;
	
	if (!settings.good()) {
//		cerr << "error: settings file " << filename << " not found.\n";
		mHasErrors = true;
		return;
	}
	while (settings.good()) {
		line++;
		char tmp[256];
		char string[256];
		settings.getline(tmp,255);
		char* ptr=tmp;
		while (*ptr!='\0' && (*ptr==' ' || *ptr=='\t')) ptr++;
		if (*ptr!='#') {
			char* key=ptr;
			while (*ptr!='\0' && *ptr!=' ' && *ptr!='\t' && *ptr!='=') ptr++;
			if (*ptr!='\0' && *ptr!='=') {
				*ptr='\0';
				ptr++;
				while (*ptr!='\0' && *ptr!='=' && 
					(*ptr==' ' || *ptr=='\t')) ptr++;
			}
			if (*ptr=='=') {
				*ptr='\0';
				ptr++;
			}
			while (*ptr!='\0' && 
				(*ptr==' ' || *ptr=='\t')) ptr++;

			char* val=ptr;

			if (*ptr=='"') {
				ptr++;
				int si=0;
				while (*ptr!='\0' && (*ptr!='"')) {
					if (*ptr=='\\') {
						ptr++;
					}
					string[si++]=*ptr;
					ptr++;			
				}
				string[si]='\0';
				val=string;
			}else{
				while (*ptr!='\0' && 
					(*ptr!=' ' || *ptr!='\t')) ptr++;
			}
					
			*ptr='\0';
			
			if (*key!='\0') {
				int err=0;
				mFound=false;

				Set(key,val);

				if (mFound==false) {
					cerr << "Line "<< line <<": error: unknown keyword '" \
						<< key << "'\n";
					mHasErrors=1;
				}
				if (err) {
					cerr << "Line "<< line <<": error: unknown value '" \
						<< val << "' for key '" << key << "'\n";
					mHasErrors=1;
				}
			}
		}
	}
	
	settings.close();
}
